#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>

#include "config.h"
#include "globals.h"
#define SYSLOG_NAMES
#ifndef _XOPEN_SOURCE_EXTENDED
#define _XOPEN_SOURCE_EXTENDED
#endif
#include "logfun.h"

#ifndef	_PATH_LOGNAME
#define	_PATH_LOGNAME "/dev/log"
#endif

static INT	LogS = -1;
static INT	LogF = -1;

static  CHAR msgBuf[8192];

static CHAR *mtime (time_t *now)
{
    static CHAR mbuf[16];

    strftime (mbuf, sizeof(mbuf), "%H:%M:%S", localtime(now));
    return mbuf;
}

static CHAR *mdtime (time_t *now)
{
    static CHAR mbuf[32];
    time_t inow;

    if (now == NULL) {
        inow = time(NULL);
        now = &inow;
    }
    strftime (mbuf, sizeof(mbuf), "%b %d %H:%M:%S", localtime(now));
    return mbuf;
}

static CHAR *gettext(CODE *cp, INT val)
{
    for (; cp->c_name != NULL; cp++)
        if (cp->c_val == val)
            return (cp->c_name);
    return ("?");
}

static INT sopen (char *USockName)
{
    INT s;
    struct sockaddr SlAddr;

    SlAddr.sa_family = AF_UNIX;
    strncpy(SlAddr.sa_data, _PATH_LOGNAME, sizeof(SlAddr.sa_data));

    if ((s = socket(PF_UNIX, SOCK_STREAM, 0)) < 0)
        return -1;

    if (!connect(s, &SlAddr, sizeof(SlAddr.sa_family)+ strlen(SlAddr.sa_data)))
        return s;


    close (s);
    return -1;
}


static CHAR *replaceMtag (char *fmt, INT err)
{
#ifdef _GNU_SOURCE
    return fmt;
#else
    CHAR *p, *s, *n, *newfmt, *errtxt;
    INT  c, errlen, fmtlen;

    for (p=fmt, c=0; *p != '\0'; p++) {
        if (*p == '%' && *(p+1) == 'm') {
            c++;
            p++;
        }
    }

    if (!c)
        return (fmt);

    errtxt = strerror(err);
    errlen = strlen(errtxt);
    fmtlen = strlen(fmt);

    newfmt = n = (CHAR *)malloc(fmtlen + (c * (errlen - 2)) + 1);

    if (newfmt == NULL) 
        return fmt;

    for (p=s=fmt; *p != '\0'; p++) {
        if (*p == '%' && *(p+1) == 'm') {
            if (p-s) {
                memcpy (n, s, p-s);
                n += (p-s);
            }
            memcpy (n, errtxt, errlen);
            n += errlen;
            s = ++p + 1;
        }
    }
    *n = '\0';
    return newfmt;
#endif
}


VOID errNoDynmem(CHAR *where)
{

    if (MsgLevel) {
        if (where == NULL) 
            doLog (LOG_ERR, "No dynamic memory");
        else
            doLog (LOG_ERR, "%s: no dynamic memory", where);
    }
}


VOID stopLog(VOID)
{
    if (LogS != -1) 
        close (LogS);

    if (LogF != -1)
        close (LogF);
	
    LogS = LogF = -1;
}


VOID doLog (INT pri, char *fmt, ...)
{
    CHAR *newfmt, mark;
    time_t now;
    va_list ap;
    INT ioc;
   
    if (pri & LOG_MASTER) {
        if (getpid() == parentPid)
            return;
        pri &= ~(LOG_MASTER);
    } 

    (VOID) time (&now);

    newfmt = replaceMtag (fmt, errno);

    va_start (ap, fmt);

    if (spMode) {
        mark = (MsgLevel <= MaxSyslog) ? '-' : ' ';
        mark = (LogFile != NULL) ? ((mark == ' ') ? '|' : '+') : mark;
        ioc = sprintf (msgBuf, "%c%8s[%d]%-7s: ",
                       mark, mtime(&now), getpid(), gettext(prioritynames, pri));

        ioc += vsprintf (&msgBuf[ioc], newfmt, ap);

        strcpy(&msgBuf[ioc++], "\n");

        write (fileno(stderr), msgBuf, ioc);
    }
    else {
        if ((LogS != -1) && (MsgLevel <= MaxSyslog)) {
            ioc = sprintf (msgBuf, "<%d>%s[%d]: ", 
                           pri|LOGFACILITY, myName, getpid());

            ioc += vsprintf (&msgBuf[ioc], newfmt, ap);

            write (LogS, msgBuf, ioc+1);
        }

        if (LogF != -1) {
            ioc = sprintf (msgBuf, "%15s [%d] %-7s: ",
                           mdtime(&now), getpid(), gettext(prioritynames, pri));

            ioc += vsprintf (&msgBuf[ioc], newfmt, ap);

            strcpy(&msgBuf[ioc++], "\n");
            write (LogF, msgBuf, ioc);
        }
    }
    if (newfmt != fmt)
        free (newfmt);
    va_end (ap);
}


VOID startLog(VOID)
{
    INT Ferr=0, Serr=0;

    stopLog();

    if (spMode)
        return;

    if (LogFile != NULL) {
        if ((LogF = open (LogFile, O_WRONLY|O_APPEND|O_NDELAY)) == -1)
            Ferr = 1;
    }
    else
        LogF = -1;

    if ((LogS = sopen (_PATH_LOGNAME)) == -1)
        Serr = 1;

    if (Serr && (LogF != -1))
        doLog (LOG_ERR, "Cannot open path to system Logger: %m");

    if (Ferr && (LogS != -1))
        doLog (LOG_ERR, "Cannot open Logfile \"%s\": %m", LogFile);

    if ((LogF != -1) && (getpid() == parentPid))
        doLog (LOG_DEBUG, "Opened Logfile \"%s\" level %d", LogFile,MsgLevel);
}

#ifdef TEST
#include "readconfig.h"

int main (int argc, char **argv)
{
    char prio[10]; char message[200];

    evalCommandLine(argc, argv, 1);

    startLog();
    
    while (1) {
        printf("Enter priority>");
        if (fgets(prio, sizeof(prio), stdin) == NULL) break;
        printf("Enter message >");
        if (fgets(message, sizeof(message), stdin) == NULL) break;
        doLog(atoi(prio), "%s", message);
    }
    return 0;
}

#endif
